/*
 * Copyright (c) 2018 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Mon Aug 20 2018
 */
#include "acmccdaqft.h"

#include <sys/time.h>
#include <err.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "daqflex/daqflex.h"

#include "mccdaqft_c_types.h"
#include "codels.h"


/* --- Task read -------------------------------------------------------- */


/** Codel daqft_init of task read.
 *
 * Triggered by mccdaqft_start.
 * Yields to mccdaqft_pause_read.
 */
genom_event
daqft_init(mccdaqft_ids *ids, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  /* default calibration */
  *ids->calibration.b = *(double [6]){0.};
  daqft_set_offset(0., 0., 0., 0., 0., 0., &ids->calibration, self);


  /* init logging */
  ids->log = malloc(sizeof(*ids->log));
  if (!ids->log) abort();
  *ids->log = (mccdaqft_log_s){
    .req = {
      .aio_fildes = -1,
      .aio_offset = 0,
      .aio_buf = ids->log->buffer,
      .aio_nbytes = 0,
      .aio_reqprio = 0,
      .aio_sigevent = { .sigev_notify = SIGEV_NONE },
      .aio_lio_opcode = LIO_NOP
    },
    .pending = false, .skipped = false,
    .decimation = 1, .missed = 0, .total = 0
  };

  return mccdaqft_pause_read;
}


/** Codel daqft_read of task read.
 *
 * Triggered by mccdaqft_read.
 * Yields to mccdaqft_export, mccdaqft_pause_read.
 */
genom_event
daqft_read(mccdevice **device, double data[6],
           const genom_context self)
{
  int s;

  if (!*device) return mccdaqft_pause_read;

  s = daqflexUSB201ReadData(*device, data, 6);
  if (s != 6) {
    warnx("could not read from device");
    daqft_disconnect_start(device, self);
    return mccdaqft_pause_read;
  }

  return mccdaqft_export;
}


/** Codel daqft_export of task read.
 *
 * Triggered by mccdaqft_export.
 * Yields to mccdaqft_read.
 */
genom_event
daqft_export(const double data[6],
             const mccdaqft_ids_calibration_s *calibration,
             mccdaqft_log_s **log, const mccdaqft_wrench *wrench,
             const genom_context self)
{
  struct timeval tv;
  double w[6];

  gettimeofday(&tv, NULL);

  /* remove bias, apply calibration matrix */
  daqft_get_data(data, calibration->G, calibration->b, w);

  /* publish */
  *wrench->data(self) = (or_wrench_estimator_state){
    .ts = { .sec = tv.tv_sec, .nsec = tv.tv_usec * 1000 },
    .intrinsic = true,

    .force = {
      ._present = true,
      ._value = { .x = w[0], .y = w[1], .z = w[1] }
    },
    .force_cov = { ._present = false },

    .torque = {
      ._present = true,
      ._value = { .x = w[2], .y = w[3], .z = w[4] }
    },
    .torque_cov = { ._present = false }
  };
  wrench->write(self);

  /* log */
  if ((*log)->req.aio_fildes >= 0) {
    (*log)->total++;
    if ((*log)->total % (*log)->decimation == 0) {
      if ((*log)->pending) {
        if (aio_error(&(*log)->req) != EINPROGRESS) {
          (*log)->pending = false;
          if (aio_return(&(*log)->req) <= 0) {
            warn("log");
            close((*log)->req.aio_fildes);
            (*log)->req.aio_fildes = -1;
          }
        } else {
          (*log)->skipped = true;
          (*log)->missed++;
        }
      }
    }

    if ((*log)->req.aio_fildes >= 0 && !(*log)->pending) {
      (*log)->req.aio_nbytes = snprintf(
        (*log)->buffer, sizeof((*log)->buffer),
        "%s" mccdaqft_log_line "\n",
        (*log)->skipped ? "\n" : "",
        (uint64_t)tv.tv_sec, (uint32_t)tv.tv_usec * 1000,
        w[0], w[1], w[2], w[3], w[4], w[5]);

      if (aio_write(&(*log)->req)) {
        warn("log");
        close((*log)->req.aio_fildes);
        (*log)->req.aio_fildes = -1;
      } else
        (*log)->pending = true;

      (*log)->skipped = false;
    }
  }

  return mccdaqft_read;
}


/** Codel daqft_stop of task read.
 *
 * Triggered by mccdaqft_stop.
 * Yields to mccdaqft_ether.
 */
genom_event
daqft_stop(mccdaqft_log_s **log, mccdevice **device,
           const genom_context self)
{
  daqft_log_stop(log, self);
  if (*log) free(*log);

  daqft_disconnect_start(device, self);

  return mccdaqft_ether;
}


/* --- Activity connect ------------------------------------------------- */

/** Codel daqft_connect_start of activity connect.
 *
 * Triggered by mccdaqft_start.
 * Yields to mccdaqft_ether.
 * Throws mccdaqft_e_nodev, mccdaqft_e_baddev, mccdaqft_e_inval.
 */
genom_event
daqft_connect_start(uint32_t rate, mccdevice **device,
                    const genom_context self)
{
  char response[64];

  /* connect to the DAQ device */
  if (*device) daqft_disconnect_start(device, self);
  *device = daqflexOpen();
  if (!*device) return mccdaqft_e_nodev(self);

  daqflexSendMessage(*device, "AISCAN:STOP", response, sizeof(response));
  daqflexFlushInputData(*device);

  /* display the serial number - just for fun for now */
  daqflexSendMessage(*device, "?DEV:MFGSER", response, sizeof(response));
  warnx("connected to device '%s'", response);


  /* init the USB-201 device: acquire channels 0 to 5 at given rate */
  daqflexUSB201Init(*device, 0, 5, rate);

  /* start an infinite loop of aquisition */
  daqflexSendMessage(*device, "AISCAN:SAMPLES=0", response, sizeof(response));
  daqflexSendMessage(*device, "AISCAN:START", response, sizeof(response));
  if (strcmp(response, "AISCAN:START")) {
    warnx("could not configure device");
    daqft_disconnect_start(device, self);
    return mccdaqft_e_baddev(self);
  }

  return mccdaqft_ether;
}


/* --- Activity disconnect ---------------------------------------------- */

/** Codel daqft_disconnect_start of activity disconnect.
 *
 * Triggered by mccdaqft_start.
 * Yields to mccdaqft_ether.
 */
genom_event
daqft_disconnect_start(mccdevice **device, const genom_context self)
{
  (void)self; /* -Wunused-parameter */
  char response[64];

  if (!*device) return mccdaqft_ether;

  daqflexSendMessage(*device, "AISCAN:STOP", response, sizeof(response));
  daqflexClose(*device);
  *device = NULL;
  warnx("disconnected from device");

  return mccdaqft_ether;
}


/* --- Activity set_zero ------------------------------------------------ */

/** Codel daqft_zero_start of activity set_zero.
 *
 * Triggered by mccdaqft_start.
 * Yields to mccdaqft_loop.
 * Throws mccdaqft_e_nodev.
 */
genom_event
daqft_zero_start(int32_t *n, const double b[6], double ib[6],
                 const genom_context self)
{
  (void)self; /* -Wunused-parameter */
  int i;

  *n = 0;
  for(i=0; i<6; i++) ib[i] = b[i];

  return mccdaqft_loop;
}

/** Codel daqft_zero_loop of activity set_zero.
 *
 * Triggered by mccdaqft_loop.
 * Yields to mccdaqft_pause_loop, mccdaqft_ether.
 * Throws mccdaqft_e_nodev.
 */
genom_event
daqft_zero_loop(const mccdevice *device, const double data[6],
                int32_t samples, const double ib[6], int32_t *n,
                double b[6], const genom_context self)
{
  int i;

  /* this is called twice per sample, because of the read + export codels */

  if (!device) return mccdaqft_e_nodev(self);
  if ((*n)++ >= 2 * samples) return mccdaqft_ether;

  for(i=0; i<6; i++) b[i] += (data[i] - ib[i]) / samples / 2;

  return mccdaqft_pause_loop;
}
